// ----------------------------------------------------------------------------
/*
 * Copyright (c) 2013 Lessmller Lasertechnik GmbH
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
// ----------------------------------------------------------------------------

#include "bootloader_module.h"

extern unsigned char v_level;
extern char text[80];

int BOOTLOADER_CMD_read_id(unsigned int *id)
{
	SDO_CMD_ cmd = SDO_CMD_(SDO_CS_INIT_UPLOAD, SDO_TRANSFER_SEGMENTED, BOOTLOADER_IDX_IDENTITY, BOOTLOADER_SUB_IDX_PART_ID, 0);

	if (SDO_CMD(BOOTLOADER_NODE_ID + BOOTLOADER_SERVER_ID, &cmd) != 0)
		return -1;

	if (SDO_CMD_REPLY(BOOTLOADER_NODE_ID + BOOTLOADER_CLIENT_ID, &cmd) != 0)
		return -1;

	*id = cmd.get_data_32();

	return SDO_check_error("Read Part ID", &cmd);
}

int BOOTLOADER_CMD_unlock(unsigned int unlock_code)
{
	SDO_CMD_ cmd = SDO_CMD_(SDO_CS_INIT_DOWNLOAD, SDO_TRANSFER_EXPEDITED, BOOTLOADER_IDX_UNLOCK, BOOTLOADER_SUB_IDX_00, 2, (unsigned char) (unlock_code & 0x00FF), (unsigned char) (unlock_code >> 8));
	
	if (SDO_CMD(BOOTLOADER_NODE_ID + BOOTLOADER_SERVER_ID, &cmd) != 0)
		return -1;

	if (SDO_CMD_REPLY(BOOTLOADER_NODE_ID + BOOTLOADER_CLIENT_ID, &cmd) != 0)
		return -1;

	return SDO_check_error("Unlock", &cmd);
}

int BOOTLOADER_CMD_erase_sector(unsigned char start_sector, unsigned char end_sector)
{
	SDO_CMD_ cmd = SDO_CMD_(SDO_CS_INIT_DOWNLOAD, SDO_TRANSFER_EXPEDITED, BOOTLOADER_IDX_ERASE_SECT, BOOTLOADER_SUB_IDX_00, 2, start_sector, end_sector);

	if (SDO_CMD(BOOTLOADER_NODE_ID + BOOTLOADER_SERVER_ID, &cmd) != 0)
		return -1;

	if (SDO_CMD_REPLY(BOOTLOADER_NODE_ID + BOOTLOADER_CLIENT_ID, &cmd) != 0)
		return -1;

	return SDO_check_error("Erase Sector", &cmd);
}

int BOOTLOADER_CMD_prepare_sector(unsigned char start_sector, unsigned char end_sector)
{
	SDO_CMD_ cmd = SDO_CMD_(SDO_CS_INIT_DOWNLOAD, SDO_TRANSFER_EXPEDITED, BOOTLOADER_IDX_PREP_SECT_WR, BOOTLOADER_SUB_IDX_00, 2, start_sector, end_sector);

	if (SDO_CMD(BOOTLOADER_NODE_ID + BOOTLOADER_SERVER_ID, &cmd) != 0)
		return -1;

	if (SDO_CMD_REPLY(BOOTLOADER_NODE_ID + BOOTLOADER_CLIENT_ID, &cmd) != 0)
		return -1;

	return SDO_check_error("Prepare Sector", &cmd);
}

int BOOTLOADER_CMD_read_boot_code(unsigned int *code)
{
	SDO_CMD_ cmd = SDO_CMD_(SDO_CS_INIT_UPLOAD, SDO_TRANSFER_EXPEDITED, BOOTLOADER_IDX_IDENTITY, BOOTLOADER_SUB_IDX_BOOTLOADER_CODE, 0);
	
	if (SDO_CMD(BOOTLOADER_NODE_ID + BOOTLOADER_SERVER_ID, &cmd) != 0)
		return-1;

	if (SDO_CMD_REPLY(BOOTLOADER_NODE_ID + BOOTLOADER_CLIENT_ID, &cmd) != 0)
		return -1;

	*code = cmd.get_data_32();

	return SDO_check_error("Read Boot ID", &cmd);
}

int BOOTLOADER_CMD_read_serial(unsigned int *serial)
{
	SDO_CMD_ cmd;
	int result = 0;

	for (unsigned char i = 0; i < 4; i++)
	{
		cmd = SDO_CMD_(SDO_CS_INIT_UPLOAD, SDO_TRANSFER_EXPEDITED, BOOTLOADER_IDX_SERIAL, BOOTLOADER_SUB_IDX_SERIAL_1 + i, 0);
	
		if (SDO_CMD(BOOTLOADER_NODE_ID + BOOTLOADER_SERVER_ID, &cmd) != 0)
			return-1;

		if (SDO_CMD_REPLY(BOOTLOADER_NODE_ID + BOOTLOADER_CLIENT_ID, &cmd) != 0)
			return -1;

		serial[i] = cmd.get_data_32();

		if (result = SDO_check_error("Read Serial", &cmd) != 0)
			return result;
	}
	return 0;
}

int BOOTLOADER_CMD_write_ram(unsigned int start_address, unsigned int size, unsigned char *data)
{
	int status = 0;

	SDO_CMD_ cmd = SDO_CMD_(SDO_CS_INIT_DOWNLOAD, SDO_TRANSFER_EXPEDITED, BOOTLOADER_IDX_RAM_WR_ADDR, BOOTLOADER_SUB_IDX_00, 4, (unsigned char) start_address, (unsigned char) (start_address >> 8), (unsigned char) (start_address >> 16), (unsigned char) (start_address >> 24));
	
	if (SDO_CMD(BOOTLOADER_NODE_ID + BOOTLOADER_SERVER_ID, &cmd) != 0)
		return -1;

	if (SDO_CMD_REPLY(BOOTLOADER_NODE_ID + BOOTLOADER_CLIENT_ID, &cmd) != 0)
		return -1;
	
	status += SDO_check_error("Write RAM Start Address", &cmd);

	cmd = SDO_CMD_(SDO_CS_INIT_DOWNLOAD, SDO_TRANSFER_SEGMENTED, BOOTLOADER_IDX_PROG_DATA, BOOTLOADER_SUB_IDX_PROG_AREA, 4,  (unsigned char) size, (unsigned char) (size >> 8), (unsigned char) (size >> 16), (unsigned char) (size >> 24));
	
	if (SDO_CMD(BOOTLOADER_NODE_ID + BOOTLOADER_SERVER_ID, &cmd) != 0)
		return -1;

	if (SDO_CMD_REPLY(BOOTLOADER_NODE_ID + BOOTLOADER_CLIENT_ID, &cmd) != 0)
		return -1;
	
	status += SDO_check_error("Write Size of Data", &cmd);

	unsigned int byte = 0, progress = 0, last_progress = 0;

	if (v_level == 1)
	{
		printf("\r\n0%%");
		for (unsigned char i = 0; i < 73; i++)
			printf(" ");
		printf("100%%\r\n[");
	}

	while ((size - byte) > 7)
	{

		if (v_level >= 4)
			printf("\r\n Uploading data [0x%04X-0x%04X] : 0x%02X 0x%02X 0x%02X 0x%02X 0x%02X 0x%02X 0x%02X", byte, byte+7, data[byte],  data[byte+1], data[byte+2], data[byte+3], data[byte+4], data[byte+5], data[byte+6]);

		if (SDO_SEG(BOOTLOADER_NODE_ID + BOOTLOADER_SERVER_ID, SDO_CS_SEG_DOWNLOAD, SDO_TRANSFER_NOT_LAST, 7, &(data[byte])) != 0)
			return -1;
	
		if (SDO_CMD_REPLY(BOOTLOADER_NODE_ID + BOOTLOADER_CLIENT_ID, &cmd) != 0)
			return -1;
		
		sprintf_s(text, 80, "Write Data Bytes [0x%04X-0x%04X]", byte, byte + 7);
		SDO_check_error(text, &cmd);

		if (v_level == 1)
		{
			progress = (unsigned int) ((77.0f / (float) size) * (byte + 7));

			if (progress > last_progress)
			{
				for (unsigned char i = 0; i < (unsigned int) ((77.0f / (float) size) * 7 + 1.0f); i++) printf("%c", 219);
				last_progress = progress;
			}
		}

		byte += 7;
	}

	if (v_level >= 4)
		printf("\r\n Uploading data[0x%04X-0x%04X] : 0x%02X 0x%02X 0x%02X 0x%02X 0x%02X 0x%02X 0x%02X\r\n", byte, byte+7, data[byte],  data[byte+1], data[byte+2], data[byte+3], data[byte+4], data[byte+5], data[byte+6]);

	if (SDO_SEG(BOOTLOADER_NODE_ID + BOOTLOADER_SERVER_ID, SDO_CS_SEG_DOWNLOAD, SDO_TRANSFER_LAST, (size-byte), &(data[byte])) != 0)
		return -1;
	
	if (SDO_CMD_REPLY(BOOTLOADER_NODE_ID + BOOTLOADER_CLIENT_ID, &cmd) != 0)
	return -1;

	if (v_level == 1)
	{
		for (unsigned char i = 1; i < (unsigned int) ((78.0f / (float) size) * (size-byte) + 0.5f); i++) printf("%c", 219);
		printf("]");
	}

	sprintf_s(text, "Write Data Bytes [0x%04X-0x%04X]", byte, size);
	status += SDO_check_error(text, &cmd);

	return status;
}

int BOOTLOADER_CMD_read_ram(unsigned int start_address, unsigned int size, unsigned char *data)
{
	int status = 0;
	unsigned char cmd_specifier = 0;
	unsigned char last = 0, reply_size = 0;
	unsigned int byte = 0, progress = 0, last_progress = 0;
 
	sprintf_s(text, "Reading Data starting at 0x%08X", start_address);
	title(text);

	SDO_CMD_ cmd = SDO_CMD_(SDO_CS_INIT_DOWNLOAD, SDO_TRANSFER_EXPEDITED, BOOTLOADER_IDX_MEM_RD_ADDR, BOOTLOADER_SUB_IDX_00, 4, (unsigned char) start_address, (unsigned char) (start_address >> 8), (unsigned char) (start_address >> 16), (unsigned char) (start_address >> 24));
	
	if (SDO_CMD(BOOTLOADER_NODE_ID + BOOTLOADER_SERVER_ID, &cmd) != 0)
		return -1;

	if (SDO_CMD_REPLY(BOOTLOADER_NODE_ID + BOOTLOADER_CLIENT_ID, &cmd) != 0)
		return -1;
	
	status += SDO_check_error("Write Start Address", &cmd);

	cmd = SDO_CMD_(SDO_CS_INIT_DOWNLOAD, SDO_TRANSFER_EXPEDITED, BOOTLOADER_IDX_MEM_RD_LEN, BOOTLOADER_SUB_IDX_00, 4, (unsigned char) size, (unsigned char) (size >> 8), (unsigned char) (size >> 16), (unsigned char) (size >> 24));
	
	if (SDO_CMD(BOOTLOADER_NODE_ID + BOOTLOADER_SERVER_ID, &cmd) != 0)
		return -1;

	if (SDO_CMD_REPLY(BOOTLOADER_NODE_ID + BOOTLOADER_CLIENT_ID, &cmd) != 0)
		return -1;
	
	status += SDO_check_error("Write Data Size", &cmd);
	
	cmd = SDO_CMD_(SDO_CS_INIT_UPLOAD, SDO_TRANSFER_SEGMENTED, BOOTLOADER_IDX_PROG_DATA, BOOTLOADER_SUB_IDX_PROG_AREA, 0);
	
	if (SDO_CMD(BOOTLOADER_NODE_ID + BOOTLOADER_SERVER_ID, &cmd) != 0)
		return -1;

	if (SDO_CMD_REPLY(BOOTLOADER_NODE_ID + BOOTLOADER_CLIENT_ID, &cmd) != 0)
		return -1;
	
	status += SDO_check_error("Read Command", &cmd);

	if (v_level == 1 && size > 8)
	{
		printf("\r\n0%%");
		for (unsigned char i = 0; i < 73; i++)
			printf(" ");
		printf("100%%\r\n[");
	}

	while ((size - byte) > 7)
	{
		if (SDO_SEG_(BOOTLOADER_NODE_ID + BOOTLOADER_SERVER_ID, SDO_CS_SEG_UPLOAD, SDO_TRANSFER_NOT_LAST, 0) != 0)
			return -1;
	
		if (SDO_SEG_REPLY(BOOTLOADER_NODE_ID + BOOTLOADER_CLIENT_ID, &last, &reply_size, &(data[byte])) != 0)
			return -1;
		
		sprintf_s(text, 80, "Read Data Bytes [0x%04X-0x%04X]", byte, byte + 7);
		SDO_check_error(text, &cmd);

		if (v_level >= 4)
			printf("\nRead Data: 0x%02X 0x%02X 0x%02X 0x%02X 0x%02X 0x%02X 0x%02X\n", data[byte], data[byte + 1], data[byte + 2], data[byte + 3], data[byte + 4], data[byte + 5], data[byte + 6]);
		
		if (v_level == 1)
		{
			progress = (unsigned int) ((78.0f / (float) size) * (byte + 7));

			if (progress > last_progress)
			{
				printf("%c", 219);
				last_progress = progress;
			}
		}
		byte += 7;
	}

	if (SDO_SEG_(BOOTLOADER_NODE_ID + BOOTLOADER_SERVER_ID, SDO_CS_SEG_UPLOAD, SDO_TRANSFER_LAST, 0) != 0)
		return -1;
	
	if (SDO_SEG_REPLY(BOOTLOADER_NODE_ID + BOOTLOADER_CLIENT_ID, &last, &reply_size, &(data[byte])) != 0)
		return -1;

	sprintf_s(text, 80, "Read Data Bytes [0x%04X-0x%04X]", byte, byte + 7);
	SDO_check_error(text, &cmd);
	
	if (v_level >= 4)
		printf("\nRead Data: 0x%02X 0x%02X 0x%02X 0x%02X 0x%02X 0x%02X 0x%02X\n", data[byte], data[byte + 1], data[byte + 2], data[byte + 3], data[byte + 4], data[byte + 5], data[byte + 6]);

	if (v_level == 1)
		printf("]\n");
	
	return status;
}

int BOOTLOADER_CMD_copy_ram_to_flash(unsigned int destination_address, unsigned int source_address, unsigned int size)// Size must be at least 256 or 512,1024,4069. Destination address should be at a 256 boundary
{
	int status = 0;

	SDO_CMD_ cmd = SDO_CMD_(SDO_CS_INIT_DOWNLOAD, SDO_TRANSFER_EXPEDITED, BOOTLOADER_IDX_CPY_RAM_MEM, BOOTLOADER_SUB_IDX_FLASH_ADDR, 4, (unsigned char) destination_address, (unsigned char) (destination_address >> 8), (unsigned char) (destination_address >> 16), (unsigned char) (destination_address >> 24));
	
	if (SDO_CMD(BOOTLOADER_NODE_ID + BOOTLOADER_SERVER_ID, &cmd) != 0)
		return -1;

	if (SDO_CMD_REPLY(BOOTLOADER_NODE_ID + BOOTLOADER_CLIENT_ID, &cmd) != 0)
		return -1;
	
	status += SDO_check_error("Write Destination Address", &cmd);

	cmd = SDO_CMD_(SDO_CS_INIT_DOWNLOAD, SDO_TRANSFER_EXPEDITED, BOOTLOADER_IDX_CPY_RAM_MEM, BOOTLOADER_SUB_IDX_RAM_ADDR, 4, (unsigned char) source_address, (unsigned char) (source_address >> 8), (unsigned char) (source_address >> 16), (unsigned char) (source_address >> 24));
	
	if (SDO_CMD(BOOTLOADER_NODE_ID + BOOTLOADER_SERVER_ID, &cmd) != 0)
		return -1;

	if (SDO_CMD_REPLY(BOOTLOADER_NODE_ID + BOOTLOADER_CLIENT_ID, &cmd) != 0)
		return -1;
	
	status += SDO_check_error("Write Source Address", &cmd);

	cmd = SDO_CMD_(SDO_CS_INIT_DOWNLOAD, SDO_TRANSFER_EXPEDITED, BOOTLOADER_IDX_CPY_RAM_MEM, BOOTLOADER_SUB_IDX_CPY_LEN, 2, (unsigned char) size, (unsigned char) (size >> 8));
	
	if (SDO_CMD(BOOTLOADER_NODE_ID + BOOTLOADER_SERVER_ID, &cmd) != 0)
		return -1;

	if (SDO_CMD_REPLY(BOOTLOADER_NODE_ID + BOOTLOADER_CLIENT_ID, &cmd) != 0)
		return -1;
	
	status += SDO_check_error("Write Copy Size", &cmd);

	return status;

}

int BOOTLOADER_CMD_compare(unsigned int flash_address, unsigned int ram_address, unsigned int size) // Size must be multiple of 4, addresses must be greater than 512 (0x200)
{
	int status = 0;

	SDO_CMD_ cmd = SDO_CMD_(SDO_CS_INIT_DOWNLOAD, SDO_TRANSFER_EXPEDITED, BOOTLOADER_IDX_CMP_MEM, BOOTLOADER_SUB_IDX_CMP_ADDR1, 4, (unsigned char) flash_address, (unsigned char) (flash_address >> 8), (unsigned char) (flash_address >> 16), (unsigned char) (flash_address >> 24));

	if (SDO_CMD(BOOTLOADER_NODE_ID + BOOTLOADER_SERVER_ID, &cmd) != 0)
		return -1;

	if (SDO_CMD_REPLY(BOOTLOADER_NODE_ID + BOOTLOADER_CLIENT_ID, &cmd) != 0)
		return -1;

	status += SDO_check_error("Write Flash Address", &cmd);

	cmd = SDO_CMD_(SDO_CS_INIT_DOWNLOAD, SDO_TRANSFER_EXPEDITED, BOOTLOADER_IDX_CMP_MEM, BOOTLOADER_SUB_IDX_CMP_ADDR2, 4, (unsigned char) ram_address, (unsigned char) (ram_address >> 8), (unsigned char) (ram_address >> 16), (unsigned char) (ram_address >> 24));

	if (SDO_CMD(BOOTLOADER_NODE_ID + BOOTLOADER_SERVER_ID, &cmd) != 0)
		return -1;

	if (SDO_CMD_REPLY(BOOTLOADER_NODE_ID + BOOTLOADER_CLIENT_ID, &cmd) != 0)
	return -1;

	status += SDO_check_error("Write RAM Address", &cmd);

	cmd = SDO_CMD_(SDO_CS_INIT_DOWNLOAD, SDO_TRANSFER_EXPEDITED, BOOTLOADER_IDX_CMP_MEM, BOOTLOADER_SUB_IDX_CMP_LEN, 2, (unsigned char) size, (unsigned char) (size >> 8));

	if (SDO_CMD(BOOTLOADER_NODE_ID + BOOTLOADER_SERVER_ID, &cmd) != 0)
		return -1;

	if (SDO_CMD_REPLY(BOOTLOADER_NODE_ID + BOOTLOADER_CLIENT_ID, &cmd) != 0)
		return -1;

	status += SDO_check_error("Write Compare Size", &cmd);

	if (cmd.data[0] == 0x0A)
	{
		cmd = SDO_CMD_(SDO_CS_INIT_UPLOAD, SDO_TRANSFER_SEGMENTED, BOOTLOADER_IDX_CMP_MEM, BOOTLOADER_SUB_IDX_CMP_OFFSET, 0);
		
		if (SDO_CMD(BOOTLOADER_NODE_ID + BOOTLOADER_SERVER_ID, &cmd) != 0)
			return -1;

		if (SDO_CMD_REPLY(BOOTLOADER_NODE_ID + BOOTLOADER_CLIENT_ID, &cmd) != 0)
			return -1;

		status += SDO_check_error("Read Offset", &cmd);

		printf("Offset: 0x%02X 0x%02X 0x%02X 0x%02X\n", cmd.data[0], cmd.data[1], cmd.data[2], cmd.data[3]);
	}

	return status;
}

int BOOTLOADER_CMD_go(unsigned int start_address)
{
	int status = 0;

	SDO_CMD_ cmd = SDO_CMD_(SDO_CS_INIT_DOWNLOAD, SDO_TRANSFER_EXPEDITED, BOOTLOADER_IDX_EXEC_ADDR, BOOTLOADER_SUB_IDX_EXEC_ADDR, 4, (unsigned char) start_address, (unsigned char) (start_address >> 8), (unsigned char) (start_address >> 16), (unsigned char) (start_address >> 24));
	
	if (SDO_CMD(BOOTLOADER_NODE_ID + BOOTLOADER_SERVER_ID, &cmd) != 0)
		return -1;

	if (SDO_CMD_REPLY(BOOTLOADER_NODE_ID + BOOTLOADER_CLIENT_ID, &cmd) != 0)
		return -1;

	status += SDO_check_error("Write Execution Address", &cmd);

	cmd = SDO_CMD_(SDO_CS_INIT_DOWNLOAD, SDO_TRANSFER_EXPEDITED, BOOTLOADER_IDX_PROG_CTRL, BOOTLOADER_SUB_IDX_PROG_CTRL, 1, (unsigned char) 0x01, (unsigned char) 0x00); // Two '...'-parameters to select correct instructor
	
	if (SDO_CMD(BOOTLOADER_NODE_ID + BOOTLOADER_SERVER_ID, &cmd) != 0)
		return -1;

	if (SDO_CMD_REPLY(BOOTLOADER_NODE_ID + BOOTLOADER_CLIENT_ID, &cmd) != 0)
		return -1;

	status += SDO_check_error("Trigger Execution", &cmd);

	return status;
}

int BOOTLOADER_CMD_reinvoke_isp(void)
{
	SDO_CMD_ cmd = SDO_CMD_(SDO_CS_INIT_DOWNLOAD, SDO_TRANSFER_EXPEDITED, 0x5001, 0x00, 1, (unsigned char) 0x00, (unsigned char) 0x00);
	if (SDO_CMD(BOOTLOADER_NODE_ID + BOOTLOADER_SERVER_ID, &cmd) != 0)
		return -1;

	if (SDO_CMD_REPLY(BOOTLOADER_NODE_ID + BOOTLOADER_CLIENT_ID, &cmd) != 0)
		return -1;

	return SDO_check_error("Reinvoke ISP", &cmd);
}