// ----------------------------------------------------------------------------
/*
 * Copyright (c) 2013 Lessmller Lasertechnik GmbH
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
// ----------------------------------------------------------------------------
#include "can_module.h"

int CAN_init(unsigned int baudrate)
{
	return TRX_init(baudrate);
}

int CAN_close(void)
{
	return TRX_close();
}

#define char_range(c)	((c>0x19) & (c<0x7F) ? c : 0x20)

int CAN_send(unsigned char buffer_num, can_message_ msg)
{
	unsigned int status;
	unsigned char tx_status;

	status = CAN_check_tx(&tx_status);

	if ((tx_status & (1 << buffer_num)) != 0)
		return -1;

#if SUPPORT_EXTENDED_CANID == 1	
	status += TRX_load_tx_buffer(buffer_num, msg.id, msg.flags.ext, msg.flags.rtr, msg.length, msg.data);
#else
	status += TRX_load_tx_buffer(buffer_num, msg.id 0, msg.flags.rtr, msg.length, msg.data);
#endif

	status += TRX_RTS(buffer_num);

	return status;
}

int CAN_recv(unsigned char buffer_num, can_message_ *msg)
{
	unsigned int status;

	for(unsigned char i = 0; i < 8; i++)
		msg->data[i] = 0x00;

#if SUPPORT_EXTENDED_CANID == 1	
	status = TRX_read_rx_buffer(buffer_num, &(msg->id), &(msg->flags.ext), &(msg->flags.rtr), &(msg->length), (unsigned char*) &(msg->data));
#else
	status = TRX_read_rx_buffer(buffer_num, &(msg->sid), 0, &(msg->flags.rtr), &(msg->length), (unsigned char*) &(msg->data));
#endif
	
	if (msg->flags.ext == 0)
		msg->id &= 0x7FF;

	status += TRX_clear_interrupts((unsigned char) (1 << buffer_num)); // RX0IF = 0x01 and RX1IF = 0x02

	return status;
}

int CAN_check_rx(signed char *buffer_num,  unsigned char *filter_num, unsigned char *ext, unsigned char *rtr)
{
	unsigned char rx_status;
	unsigned int status;

	status = TRX_read_rx_status(&rx_status);

	*buffer_num = ((rx_status & 0xC0) >> 6) - 1;
	
	if (filter_num != NULL)
		*filter_num = rx_status & 0x07;

	if (ext != NULL)
		*ext = (rx_status & 0x10) >> 4;

	if (rtr != NULL)
		*rtr = (rx_status & 0x08) >> 3;

	return status;
}

int CAN_check_tx(unsigned char *buffer_num)
{
	unsigned char tx_status;
	unsigned int status;

	status = TRX_read_status(&tx_status);

	*buffer_num  = (tx_status & 0x04) >> 2;
	*buffer_num += (tx_status & 0x10) >> 3;
	*buffer_num += (tx_status & 0x40) >> 4;

	return status;
}

int CAN_set_filter(unsigned char filter_num, unsigned char mask_num, can_filter_ *flt)
{
	unsigned int status = 0;
	
	status += TRX_switch_mode(OPMODE_CONFIG);

	status += TRX_set_filter_id(filter_num, flt->id, flt->ext);
	status += TRX_set_mask_id(mask_num, flt->mask, flt->ext);
	
	switch (filter_num)
	{
		case 0: status += TRX_bit_modify(RXF0SIDL, 1 << EXIDE , flt->ext << EXIDE); break;
		case 1: status += TRX_bit_modify(RXF1SIDL, 1 << EXIDE , flt->ext << EXIDE); break;
		case 2: status += TRX_bit_modify(RXF2SIDL, 1 << EXIDE , flt->ext << EXIDE); break;
		case 3: status += TRX_bit_modify(RXF3SIDL, 1 << EXIDE , flt->ext << EXIDE); break;
		case 4: status += TRX_bit_modify(RXF4SIDL, 1 << EXIDE , flt->ext << EXIDE); break;
		default:
		case 5: status += TRX_bit_modify(RXF5SIDL, 1 << EXIDE , flt->ext << EXIDE); break;
	}
	
	status += TRX_switch_mode(OPMODE_NORMAL);

	return status;
}

int CAN_enable_filter(unsigned char buffer_num, unsigned char filter_num, unsigned char mode)
{
	unsigned int status = 0;

	status += TRX_switch_mode(OPMODE_CONFIG);

	status += TRX_set_filter_mode(buffer_num, mode);

	switch (mode)
	{
		case CAN_FILTER_ONLY_EXT : status = TRX_select_filter(buffer_num, filter_num, 1); break;
		case CAN_FILTER_ONLY_STD : status = TRX_select_filter(buffer_num, filter_num, 0); break;
	}
	
	status += TRX_switch_mode(OPMODE_NORMAL);

	return status;
}

int CAN_reset(unsigned int baudrate)
{
	return 0;
}

int CAN_clear(void)
{
	return TRX_clear_interrupts(0xFF);
}

int CAN_status(can_status_ *status)
{
	return TRX_status(status);
}