// ----------------------------------------------------------------------------
/*
 * Copyright (c) 2013 Lessmller Lasertechnik GmbH
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
// ----------------------------------------------------------------------------
#include "isp_module.h"

extern unsigned char v_level;
extern char text[80];

const unsigned char system_reset[16] = {0x01,0x4A,0x02,0x4B,0xDA,0x60,0xC0,0x46,0x04,0x00,0xFA,0x05,0x00,0xED,0x00,0xE0};

unsigned char connected = 0;

int ISP_connect(unsigned int baudrate, unsigned int part_id)
{
	unsigned int lpc_part_id = 0;
	unsigned char t = 0;

	title("Connecting to CAN Transceiver");
	if (CAN_init(baudrate) != 0)
		return -1;
	
	if (v_level >= 1)
		title("Searching LPC");

	do {
		Sleep(500);
		if (v_level >= 1)
			printf(".");
	}
	while ((BOOTLOADER_CMD_read_id(&lpc_part_id) != 0) && (t++ < ISP_SEARCH_TIMEOUT));
	
	if (t == ISP_SEARCH_TIMEOUT)
		return -4;
	
	if (v_level == 1)
		printf("OK");
	
	if (v_level >= 2)
	{
		sprintf_s(text, "Part ID read : %08X \t Expected : %08X", lpc_part_id, part_id);
		title(text);
	}
	
	if (part_id != lpc_part_id)
		return -5;
	
	if (v_level >= 1)
		title("Correct LPC found!");

	connected = 1;

	return 0;
}

int ISP_reinvoke_isp(void)
{
	if (connected != 1)
		return -13;

	if (v_level >= 2)
		title("Reinvoking ISP");
	
	if (BOOTLOADER_CMD_reinvoke_isp() != 0)
		return -14;
	
	return 0;
}

int ISP_write_firmware(char *filename, unsigned int offset)
{
	unsigned char data[FLASH_SIZE];
	unsigned int filesize = 0, copysize = 0;
	unsigned char end_sector = 0, start_sector = 0;

	if (connected != 1)
		return -13;

	switch (load_bin_file(filename, &filesize, data))
	{
		case -1 : return -2;
		case -2 : return -3;
	}

	if (v_level >= 1)
	{
		sprintf_s(text, "File '%s' (%d bytes) loaded!", filename, filesize);
		title(text);
	}

	// Calculate sectors needed for data. Data before start_address is not written and does not contribute to size.
	start_sector =                 (unsigned char) ((float) offset  / (float) FLASH_SECTOR_SIZE);
	end_sector   =  start_sector + (unsigned char) ((float)filesize / (float) FLASH_SECTOR_SIZE);
	
	if (v_level >= 1)
	{
		sprintf_s(text, "Code will be written in Sectors %d to %d.", start_sector, end_sector);
		title(text);
	}

	if (BOOTLOADER_CMD_prepare_sector(start_sector, end_sector) != 0)
		return -7;

	if (v_level >= 2)
		title("Sector(s) prepared for erasing!");
	
	if (BOOTLOADER_CMD_erase_sector(start_sector, end_sector) != 0)
		return -8;

	if (v_level >= 2)
		title("Sector(s) erased!");

	title("Starting Firmware download");

	for (unsigned char sector = start_sector; sector <= end_sector; sector++)
	{
	
		if (v_level >= 1)
		{
			sprintf_s(text, "Writing Data for Sector %d to RAM.", sector);
			title(text);
		}

		if ((unsigned int) ((sector + 1) * FLASH_SECTOR_SIZE) < filesize)
			copysize = FLASH_SECTOR_SIZE;
		else
			copysize = filesize - (sector * FLASH_SECTOR_SIZE);
		
		if (BOOTLOADER_CMD_write_ram(RAM_START_ADDRESS, copysize, &(data[offset + sector * FLASH_SECTOR_SIZE])) !=0)
			return -9;

		if (v_level >= 2)
			title("Data written to RAM!");

		if (BOOTLOADER_CMD_prepare_sector(sector, sector) != 0)
			return -10;

		if (v_level >= 2)
			title("Sector prepared for writing!");

		if (BOOTLOADER_CMD_copy_ram_to_flash(FLASH_SECTOR(sector), RAM_START_ADDRESS, FLASH_SECTOR_SIZE) != 0)
			return -11;

		if (v_level >= 1)
			title("Data copied from RAM to Flash!");

		if (BOOTLOADER_CMD_compare(FLASH_SECTOR(sector), RAM_START_ADDRESS, FLASH_SECTOR_SIZE) != 0)
			return -15;

		if (v_level >=2)
			title("Sector compared!");
	}

	return 0;
}

int ISP_unlock(unsigned int unlock_code)
{
	if (connected != 1)
		return -13;

	if (v_level >= 2)
	{
		sprintf(text, "Unlocking with code 0x%04X", unlock_code);
		title(text);
	}

	if (BOOTLOADER_CMD_unlock(unlock_code) != 0)
		return -6;
	
	if (v_level >= 2)
		title("Unlock successfull!");

	return 0;
}

int ISP_disconnect(void)
{
	if (connected != 1)
		return -13;

	if (CAN_close() != 0)
		return -20;

	return 0;
}

int ISP_reset(void)
{
	unsigned char t = 0;
	unsigned int part_id = 0;

	if (connected != 1)
		return -13;

	title("Writing System Reset Sequence to RAM");

	if (BOOTLOADER_CMD_write_ram(RAM_START_ADDRESS, 16, (unsigned char*) &system_reset) !=0)
		return -9;

	if (v_level >= 2)
			title("Data written to RAM!");

	if (BOOTLOADER_CMD_go(RAM_START_ADDRESS) != 0)
		return -12;
	
	if (v_level >= 1)
		title("System Reset Sequence executed!");

	return 0;
}

int ISP_run(unsigned int start_address)
{
	unsigned char t = 0;
	unsigned int part_id = 0;

	if (connected != 1)
		return -13;

	sprintf_s(text, "Starting User Code from Address 0x%08X", start_address);
	title(text);

	if (BOOTLOADER_CMD_go(start_address) != 0)
		return -12;
	
	if (v_level >= 1)
		title("Execution of User Programm initiated!");

	return 0;
}

int ISP_erase_chip(void)
{
	unsigned int filesize = 0, copysize = 0;
	unsigned int part_id = 0;
	unsigned char t = 0;

	if (connected != 1)
		return -13;

	t =  (unsigned char) (FLASH_SIZE / FLASH_SECTOR_SIZE);

	title("Erasing sectors");

	if (BOOTLOADER_CMD_prepare_sector(START_SECTOR, t-1) != 0)
		return -7;

	if (v_level >= 2)
		title("Sectors prepared for erasing!");
	
	if (BOOTLOADER_CMD_erase_sector(START_SECTOR, t-1) != 0)
		return -8;

	if (v_level >= 2)
		title("Sectors erased!");

	return 0;
}

int ISP_verify(char *filename)
{
	unsigned char data[FLASH_SIZE];
	unsigned char read_data[FLASH_SIZE];
	unsigned int filesize = 0;
	unsigned int copysize = 0;
	unsigned int count = 0;
	unsigned char t = 0;

	if (connected != 1)
		return -13;

	switch (load_bin_file(filename, &filesize, data))
	{
		case -1 : return -2;
		case -2 : return -3;
	}

	if (v_level >= 1)
	{
		sprintf_s(text, "File '%s' (%d bytes) loaded!", filename, filesize);
		title(text);
	}

	t =  (unsigned char) (((float) filesize / (float) FLASH_SECTOR_SIZE));

	for (unsigned char sector = 0; sector <= t; sector++)
	{
		if (sector == 0)
			count = 512;
		else
			count = 0;

		if ((unsigned int) ((sector + 1) * FLASH_SECTOR_SIZE) < filesize)
			copysize = FLASH_SECTOR_SIZE;
		else
			copysize = filesize - (sector * FLASH_SECTOR_SIZE);

		copysize += copysize % 4;

		if (v_level >= 1)
		{
			title("Starting verify!");
		}
			
		
		if (BOOTLOADER_CMD_read_ram(FLASH_SECTOR(sector), copysize, read_data) != 0)
			return -17;
		
		while (count < copysize)
		{
			if (data[(sector * FLASH_SECTOR_SIZE) + count] != read_data[count])
			{
				if (v_level >= 2)
					printf("Not equal! Byte: %d data: %02X, read_data: %02X\n", count, data[(sector * FLASH_SECTOR_SIZE) + count], read_data[count]);
				
				if (v_level == 1)
					printf("\nNot equal! Byte %d\n", count);

				return -16;
			}
			if (v_level >= 3)
				printf("Byte: %d data: %02X, read_data: %02X\n", count, data[(sector * FLASH_SECTOR_SIZE) + count], read_data[count]);
			
			count++;
		}	
		if (v_level >= 1)
			title("Sector verified!");
	}
	return 0;
}

int ISP_verify_byte(char *filename)
{
	unsigned char data[FLASH_SIZE];
	unsigned char read_data[16];
	unsigned int filesize = 0;
	unsigned int copysize = 0;
	unsigned int count = 0;
	unsigned int byte;
	unsigned int start_address;
	unsigned char t = 0;

	if (connected != 1)
		return -13;

	switch (load_bin_file(filename, &filesize, data))
	{
		case -1 : return -2;
		case -2 : return -3;
	}

	if (v_level >= 1)
	{
		sprintf_s(text, "File '%s' (%d bytes) loaded!", filename, filesize);
		title(text);
	}

	t =  (unsigned char) (((float) filesize / (float) FLASH_SECTOR_SIZE));

	for (unsigned char sector = 0; sector <= t; sector++)
	{
		if (sector == 0)
		{
			count = 512;
			start_address = 0x200;
		}
		else
		{
			count = 0;
			start_address = FLASH_SECTOR(sector);
		}

		if ((unsigned int) ((sector + 1) * FLASH_SECTOR_SIZE) < filesize)
			copysize = FLASH_SECTOR_SIZE;
		else
			copysize = filesize - (sector * FLASH_SECTOR_SIZE);
	
		if (v_level >= 1)
		title("Starting verification!");

		while (count < copysize)
		{
			if (BOOTLOADER_CMD_read_ram(start_address, 8, read_data) != 0)
				return -17;
			
			for (byte = 0; byte < 8; byte++)
			{
				if (data[(sector * FLASH_SECTOR_SIZE) + count] != read_data[byte])
				{
					if (v_level >= 2)
						printf("Not equal at Byte: %d data: %02X, read_data: %02X\n", count, data[(sector * FLASH_SECTOR_SIZE) + count], read_data[byte]);
					
					if (v_level == 1)
						printf("\nNot equal at Byte %d\n", count);
					
					return -16;
				}
				if (v_level >= 3)
					printf("Byte: %d data: %02X, read_data: %02X\n", count, data[(sector * FLASH_SECTOR_SIZE) + count], read_data[byte]);
				count++;
			}
			start_address += 8;
		}
		if (v_level >= 1)
			title("Sector verified!");
	}
	return 0;
}


int load_bin_file(char *filename, unsigned int *filesize, unsigned char *data)
{
	int fp, i = 0;
	int checksum = 0;

	if ((fp = _sopen(filename, O_RDONLY | O_BINARY, SH_DENYNO)) == NULL)
		return -1; 

	if ((*filesize = _lseek(fp, 0L, SEEK_END)) < 0)
		return -1;

	if (*filesize > FLASH_SIZE)
		return -2;

	_lseek(fp, 0L, SEEK_SET);

	memset(data, 0, FLASH_SIZE);	// Clear Memory

	while (_read(fp, &data[i++], 1) != EOF);
	
	_close(fp);

	*filesize += *filesize % 4;	// Pad size to multiple of 4

	for (i = 0; i < 4; i++)
		data[i + 0x1C] = 0;

    // Calculate a native checksum of the little endian vector table:
    for (i = 0; i < (4 * 8);)
	{
        checksum += data[i++];
        checksum += data[i++] << 8;
        checksum += data[i++] << 16;
        checksum += data[i++] << 24;
    }

    checksum = (int) (0 - checksum);

	for (i = 0; i < 4; i++)
        data[i + 0x1C] = (unsigned char)(checksum >> (8 * i));

	return 0;
}


