// ----------------------------------------------------------------------------
/*
 * Copyright (c) 2013 Lessmller Lasertechnik GmbH
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
// ----------------------------------------------------------------------------

#include "mcp2515_module.h"


/* TRX_init
 * 
 * This Function initialises the MCP2515 and configures some basic settings.
 * The Baudrate is set to the value given by the parameter,
 * all Filters are disabled, i.e. all messages are received.
 */
int TRX_init(unsigned int baudrate)
{
	unsigned int status = 0;

	if (status = SPI_init() != 0)							// Initialise the SPI Communication
		return status;

	status += TRX_reset();									// Perfom one Reset to ensure the Chip is in Configuration Mode
	
	status += TRX_set_baudrate(baudrate);					// Set CAN Baudrate

	status += TRX_set_filter_mode(0, RXM_RCV_ALL);			// Set Filter for Buffer 0 to "Receive All" (= Ignore Filters)
	status += TRX_set_filter_mode(1, RXM_RCV_ALL);			// Set Filter for Buffer 1 to "Receive All" (= Ignore Filters)

	for (unsigned char i = 0; i < 6; i++)					// Loop through all 6 Filters
	{
		status += TRX_set_filter_id(i, 0x000, 0);				// Set StandardID to 0x000
#if	SUPPORT_EXTENDED_CANID == 1
		status += TRX_set_filter_id(i, 0x000, 1);				// Set ExtendedID to 0x000
		status += TRX_select_filter(0, i, EXIDE_SET);		// Configure each Filter to apply to Extended IDs
#else
		status += TRX_select_filter(0, i, EXIDE_RESET);		// Configure each Filter to apply only to Standard IDs
#endif
	}

	for (unsigned char i = 0; i < 2; i++)					// Loop through all 2 Masks
	{
		status += TRX_set_mask_id(i, 0x000, 1);				// Set Mask for Standard & Ext ID to 0x000
	}
	
	status += TRX_bit_modify(BFPCTRL, 0xFF/*(B0BFE | B1BFE | B0BFM | B1BFM)*/, 0x00);	// Disable the RXnBF Pins and set their mode to Digital Output
	status += TRX_bit_modify(TXRTSCTRL, 0xFF/*(TXB0RTS | TXB1RTS | TXB2RTS)*/, 0x00);	// Set the TXnRTS Pins to Digital Input Mode (= Request-to-Send disabled)
    
	TRX_disable_interrupts(MERRE | WAKIE | ERRIE | TX2IE | TX1IE | TX0IE);		// Disable all interrupts
	TRX_enable_interrupts(RX1IE | RX0IE);										// except the "Receive Buffer n Full" Interrupt

	TRX_clear_interrupts(0xFF);													// Clear all Interrupts

	status += TRX_switch_mode(OPMODE_NORMAL);									// Switch back to normal Operating Mode
		
	return status;
}


/* TRX_close
 * End Communication with the MCP2515 by closing the SPI Channel
 */
int TRX_close(void)
{
	return SPI_close_channel();
}

inline unsigned int TRX_read_interrupt(unsigned char * interrupt)
{
	return TRX_read_register(CANINTF, 1, interrupt);	// Read the Interrupt Flags register
}

inline unsigned int TRX_enable_interrupts(unsigned char interrupts)
{
	return TRX_bit_modify(CANINTE, interrupts, 0xFF);	// Set bits in the Interrupt Enable register to '1' according to provided mask
}

inline unsigned int TRX_clear_interrupts(unsigned char interrupts)
{
	unsigned int status;

	status = TRX_bit_modify(CANINTF, interrupts, 0x00);		// Set all bits in the Interrupt Enable register to '0'
	status += TRX_read_interrupt(&interrupts);				// Found to be necessary in order to set the correct value

	return status;
}

inline unsigned int TRX_disable_interrupts(unsigned char interrupts)
{
	return TRX_bit_modify(CANINTE, interrupts, 0x00);		// Set bits in the Interrupt Enable register to '0' according to provided mask
}

inline unsigned int TRX_read_mode(unsigned char * mode)
{
	unsigned int status = TRX_read_register(CANCTRL, 1, mode);	// Read CAN Control register
	*mode &= REQOP;												// Mask OPMODE bits
	return status;
}

/* TRX_switch_mode
 * Change the Operating Mode of the MCP2515
 */
unsigned int TRX_switch_mode(unsigned char mode)
{
	unsigned int status;
	unsigned char new_mode = 0;
	unsigned int timeout = 0;

	SPI_reset_CS();									// Reset CS to ensure the next byte is interpreted as a command

	status = TRX_bit_modify(CANCTRL, REQOP, mode);	// Set the OPMode Bits in the CAN Control register

	Sleep(10);										// Wait for Mode Change to be completed

	while (new_mode != mode)						// Check if the Mode has been changed
	{
		if (timeout++ > MCP_TIMEOUT)				// If Timeout has been reached, exit the function with errorcode
		{
		#ifdef _DEBUG
			printf("\r\nModeSwitch Timeout\r\n");
		#endif
			return -1;
		}

		TRX_read_mode(&new_mode);					// Read the current Mode
		Sleep(1);
	}

	return status;
}

inline unsigned int TRX_abort_transmission(void)
{
	unsigned int status = TRX_bit_modify(CANCTRL, ABAT, ABAT);	// Set the 'Abort Transmission' Bit in the CAN Control register. This aborts the current message Transmission
	return (status + TRX_bit_modify(CANCTRL, ABAT, ~ABAT));		// Reset the bit
}

inline unsigned int TRX_retry_transmission(bool enable)
{
	if (enable)
		return TRX_bit_modify(CANCTRL, OSM, OSM_ENABLED);		// Set the 'One Shot' Bit in the CAN Control Register. This enables Message Retransmission until it has been received.
	else
		return TRX_bit_modify(CANCTRL, OSM, ~OSM_ENABLED);		// Unset the 'One Shot' Bit in the CAN Control Register. Messages will only be sent once
}

/* TRX_set_baudrate 
 * Configures the Timing and the Baudrate of the MCP2515
 */
int TRX_set_baudrate(unsigned int baudrate)
{
	unsigned char data[3];

	data[0] = SOF_DISABLED | WAKFIL_DISABLED | PHSEG2_LENGTH;			/*CNF3*/
	data[1] = BTLMODE_CNF3 | SMPL_1X | PHSEG1_LENGTH | TPROP_LENGTH;	/*CNF2*/
	data[2] = TSJW_LENGTH | BR_PRESCALER(baudrate);						/*CNF1*/

	return TRX_write_register(CNF3, 3, data);
}

unsigned int TRX_set_filter_mode(unsigned char buffer_num, unsigned char filter_mode)
{
	if (buffer_num == 0)
		return TRX_bit_modify(RXB0CTRL, RXM, filter_mode);
	else
		return TRX_bit_modify(RXB1CTRL, RXM, filter_mode);
}

unsigned int TRX_select_filter(unsigned char buffer_num, unsigned char filter_num, unsigned char extended)
{
	unsigned int status;

	if ((buffer_num == 0) & (filter_num < 2))
		status = TRX_bit_modify(RXB0CTRL, FILHIT0, filter_num);
	else
		status = TRX_bit_modify(RXB1CTRL, FILHIT25, filter_num);

	switch (filter_num)
	{
		case 0: status += TRX_bit_modify(RXF0SIDL, 1 << EXIDE , extended << EXIDE); break;
		case 1: status += TRX_bit_modify(RXF1SIDL, 1 << EXIDE , extended << EXIDE); break;
		case 2: status += TRX_bit_modify(RXF2SIDL, 1 << EXIDE , extended << EXIDE); break;
		case 3: status += TRX_bit_modify(RXF3SIDL, 1 << EXIDE , extended << EXIDE); break;
		case 4: status += TRX_bit_modify(RXF4SIDL, 1 << EXIDE , extended << EXIDE); break;
		default:
		case 5: status += TRX_bit_modify(RXF5SIDL, 1 << EXIDE , extended << EXIDE); break;
	}
	
	return status;
}

unsigned int TRX_set_filter_id(unsigned char filter_num, unsigned int id, unsigned char extended)
{
	unsigned int status;
	unsigned int offset = filter_num * 4;
	
	if (filter_num > 2)
		offset = 0x10 + (filter_num - 3) * 4;

	// Standard ID
	status  = TRX_bit_modify(RXF0SIDH + offset, 0xFF, (unsigned char) (id >> 3));
	status += TRX_bit_modify(RXF0SIDL + offset, 0xE0, (unsigned char) (id << 5));

	// Extended ID
#if SUPPORT_EXTENDED_CANID == 1
	if (extended)
	{
		status += TRX_bit_modify(RXF0SIDL + offset, 0x03, (unsigned char) (id >> 27));
		status += TRX_bit_modify(RXF0EID8 + offset, 0xFF, (unsigned char) (id >> 19));
		status += TRX_bit_modify(RXF0EID0 + offset, 0xFF, (unsigned char) (id >> 11));
	}
#endif
	return status;
}

unsigned int TRX_set_mask_id(unsigned char mask_num, unsigned int id, unsigned char extended)
{
	unsigned int status;

	unsigned int offset = mask_num * 4;
	
	status  = TRX_bit_modify(RXM0SIDH + offset, 0xFF, (unsigned char) (id >> 3));
	status += TRX_bit_modify(RXM0SIDL + offset, 0xE0, (unsigned char) (id << 5));
#if SUPPORT_EXTENDED_CANID == 1
	if (extended)
	{
		status += TRX_bit_modify(RXM0SIDL + offset, 0x03, (unsigned char) (id >> 27));
		status += TRX_bit_modify(RXM0EID8 + offset, 0xFF, (unsigned char) (id >> 19));
		status += TRX_bit_modify(RXM0EID0 + offset, 0xFF, (unsigned char) (id >> 11));
	}
#endif
	return status;
}

unsigned int TRX_read_rx_buffer(unsigned char buffer_num, unsigned int *id, unsigned char *ext, unsigned char *rtr, unsigned char *size, unsigned char *data)
{
	unsigned int status;
	unsigned char buffer[5];

	buffer[0] = CAN_RD_RX_BUF | (buffer_num << 2);
	status  = SPI_write_data(1, buffer);
	
	status += SPI_read_data(5, buffer);
	status += SPI_read_data(8, data);

	*id = (unsigned int) (buffer[0] << 3) + (unsigned int) (buffer[1] >> 5);
#if	SUPPORT_EXTENDED_CANID == 1
	*id += (unsigned int) ((buffer[1] & 0x03) << 27) + (unsigned int) (buffer[2] << 19) + (unsigned int) (buffer[3] << 11);
#else
	*eid = 0x0000;
#endif
	*ext  = (unsigned char) (buffer[1] & 0x08) >> 3;
	
	if (*ext == 1)
		*rtr  = (unsigned char) (buffer[4] & 0x40) >> 6;
	else
		*rtr  = (unsigned char) (buffer[1] & 0x10) >> 4;

	*size = (unsigned char) (buffer[4] & 0x0F);

	if (*size > 8) *size = 8;

	SPI_reset_CS();

	return status;
}

unsigned int TRX_load_tx_buffer(unsigned char buffer_num, unsigned int id, unsigned char ext, unsigned char rtr, unsigned char size, unsigned char * data)
{
	unsigned int status;
	unsigned char buffer[6];

	buffer[0] = (unsigned char) (CAN_LD_TX_BUF | (buffer_num * 2));

	buffer[1] = (unsigned char) (id >> 3);
#if	SUPPORT_EXTENDED_CANID == 1
	buffer[2] = (unsigned char) ((id << 5) | ((ext & 0x01) << 3) | ((id >> 27) & 0x03));
	buffer[3] = (unsigned char) ((id >> 19) & 0xFF);
	buffer[4] = (unsigned char) ((id >> 11) & 0xFF);
	buffer[5] = (unsigned char) (((rtr & 0x01) << 6) | (size & 0x0F));
#else
	buffer[2] = (unsigned char) (id << 5);
	buffer[3] = (unsigned char) 0x00;
	buffer[4] = (unsigned char) 0x00;
	buffer[5] = (unsigned char) (((rtr & 0x01) << 6) | (size & 0x0F));
#endif
	status  = SPI_write_data(6, buffer);
	status += SPI_write_data(size, data);

	SPI_reset_CS();
	
	return status;
}

unsigned int TRX_RTS(unsigned char buffer_num)
{
	unsigned int status;
	unsigned char buffer;

	buffer = CAN_RTS | (1 << buffer_num);
	
	status = SPI_write_data(1, &buffer);

	SPI_reset_CS();

	return status;
}

unsigned int TRX_read_rx_status(unsigned char *rx_status)
{
	unsigned int status;
	unsigned char buffer;
	
	buffer = CAN_RX_STATUS;

	status = SPI_write_data(1, &buffer);

	status += SPI_read_data(1, rx_status);

	SPI_reset_CS();

	return status;
}

unsigned int TRX_read_status(unsigned char *status)
{
	unsigned int status_;
	unsigned char buffer;
	
	buffer = CAN_RD_STATUS;

	status_ = SPI_write_data(1, &buffer);

	status_ += SPI_read_data(1, status);

	SPI_reset_CS();

	return status_;
}

unsigned int TRX_reset(void)
{
	unsigned int status;
	unsigned char buffer;

	buffer = SPI_RESET;

	status = SPI_write_data(1, &buffer);

	Sleep(1);
	
	SPI_reset_CS();

	Sleep(10);
	return status;
}

unsigned int TRX_write_register(unsigned char address, unsigned int size, unsigned char *data)
{
	unsigned int status;
	unsigned char buffer[2];
	
	buffer[0] = SPI_WRITE;
	buffer[1] = address;
	
	status = SPI_write_data(2, buffer);
	
	status += SPI_write_data(size, data);

	SPI_reset_CS();

	return status;
}

unsigned int TRX_read_register(unsigned char address, unsigned int size, unsigned char *data)
{
	unsigned int status;
	unsigned char buffer[2];

	buffer[0] = SPI_READ;
	buffer[1] = address;
	
	status = SPI_write_data(2, buffer);

	status += SPI_read_data(size, data);

	SPI_reset_CS();

	return status;
}

unsigned int TRX_bit_modify(unsigned char address, unsigned char mask, unsigned char data)
{
	unsigned char buffer[4];

	buffer[0] = SPI_BIT_MODIFY;
	buffer[1] = address;
	buffer[2] = mask;
	buffer[3] = data;

	return SPI_write_data(4, buffer);

	SPI_reset_CS();
}

int TRX_status(trx_status_ *status)
{
	unsigned int status_ = 0;

	status_ += TRX_read_rx_status((unsigned char*) &(status->rx_status));
	status_ += TRX_read_status((unsigned char*) &(status->int_status));
	status_ += TRX_read_register(EFLG, 1, (unsigned char*) &(status->error));
	status_ += TRX_read_register(CANSTAT, 1, (unsigned char*) &(status->config));
	status_ += TRX_read_register(CANINTF, 1, (unsigned char*) &(status->int_flags));
	status_ += TRX_read_register(CANINTE, 1, (unsigned char*) &(status->int_enable));
	status_ += TRX_read_register(REC, 1, (unsigned char*) &(status->rec));
	status_ += TRX_read_register(TEC, 1, (unsigned char*) &(status->tec));

	status_ += TRX_read_register(CNF3, 3, (unsigned char*) &(status->conf));
	
	return status_;
}