// ----------------------------------------------------------------------------
/*
 * Copyright (c) 2013 Lessmller Lasertechnik GmbH
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
// ----------------------------------------------------------------------------
#ifndef _MCP2515_MODULE_
#define _MCP2515_MODULE_

/* MCP 2515 Module 
 * 
 * This Module provides functions necessary for configuring and using the MCP2515 CAN Transceiver.
 *
 */

#include "mcp2515_defs.h"				// Registers and Bit-Positions
#include "spi_module.h"

#define MCP_TIMEOUT				500		// Number of retries (every millisecond) to switch Mode of MCP2515
#define	SUPPORT_EXTENDED_CANID	1		// Support 29-bit CAN IDs

/* CAN Sampling Timing */
#define TRX_TSJW_TQ				1
#define TRX_TPROP_TQ			1

#define TRX_TPSH1_TQ			3
#define TRX_TPSH2_TQ			3
#define TRX_DELAY_TQ			2		// Not defined in Datasheet, approx. 1-2TQ

/* Calculations for Register Values */

#define TSJW_LENGTH				(TRX_TSJW_TQ - 1) << SJW0
#define TPROP_LENGTH			(TRX_TPROP_TQ - 1) << PRSEG0
#define PHSEG1_LENGTH			(TRX_TPSH1_TQ - 1) << PHSEG10
#define PHSEG2_LENGTH			(TRX_TPSH2_TQ - 1) << PHSEG20

/* Plausibility Checks */

#if (TRX_TPROP_TQ + TRX_TPSH1_TQ < TRX_TPSH2_TQ)
	#warning PropSeg + PS1 must be greater than or equal to PS2
#endif

#if (TRX_TPSH2_TQ < TRX_TSJW_TQ)
	#warning PS2 must be greater than SJW
#endif

#if (TRX_TPROP_TQ + TRX_TPSH1_TQ < TRX_DELAY_TQ)
	#warning PS2 must be greater than or equal to T_Delay
#endif

/* Baudrate Prescaler Calculation */

#define TRX_TBIT_TQ				(1 + TRX_TPROP_TQ + TRX_TPSH1_TQ + TRX_TPSH2_TQ)
#define TRX_FOSC				16000000
#define BR_PRESCALER(BAUDRATE)	(unsigned char)((unsigned char)(TRX_FOSC / ( BAUDRATE * TRX_TBIT_TQ)) / 2) - 1	// Calculate Precscaler Register

/* Defines for CAN Filter matching criteria */
#define CAN_FILTER_ALL_MSG	11
#define CAN_FILTER_ONLY_STD 01
#define CAN_FILTER_ONLY_EXT 10
#define CAN_FILTER_ALL_FRM	00

#pragma pack(push,1)		// Pack bitfields in one byte

/* Data Structures for easy access to important registers */
typedef struct
{
	unsigned char rx0if  : 1;
	unsigned char rx1if  : 1;
	unsigned char tx0req : 1;
	unsigned char tx0if  : 1;
	unsigned char tx1req : 1;
	unsigned char tx1if  : 1;
	unsigned char tx2req : 1;
	unsigned char tx2if  : 1;
} trx_int_status_;	// Data provided by the "READ STATUS" Command

typedef struct {
	unsigned char msg_buf  : 2;
	unsigned char rsv	   : 1;		// Placeholder for unused bit
	unsigned char msg_type : 2;
	unsigned char flt_mat  : 3;
} trx_rx_status_;	// Data provided by the "READ RX STATUS" Command

typedef struct {
	unsigned char rx1ovr : 1;
	unsigned char rx0ovr : 1;
	unsigned char txbo   : 1;
	unsigned char txep   : 1;
	unsigned char rxep   : 1;
	unsigned char txwar  : 1;
	unsigned char rxwar  : 1;
	unsigned char wwarn  : 1;
} trx_error_;	// EFLG Register	(Error Flags)

typedef struct {
	unsigned char opmod : 3;
	unsigned char rsv   : 1;		// Placeholder for unused bit
	unsigned char icod  : 3;
	unsigned char rsv_  : 1;		// Placeholder for unused bit
} trx_config_;	// CANSTAT Register	(MCP2515 Status)

typedef struct {
	unsigned char merrf : 1;
	unsigned char wakif : 1;
	unsigned char errif : 1;
	unsigned char tx2if : 1;
	unsigned char tx1if : 1;
	unsigned char tx0if : 1;
	unsigned char rx1if : 1;
	unsigned char rx0if : 1;
} trx_int_flags_;	// CANINTF Register	(Interrupt Flags)

typedef struct {
	unsigned char merre : 1;
	unsigned char wakie : 1;
	unsigned char errie : 1;
	unsigned char tx2ie : 1;
	unsigned char tx1ie : 1;
	unsigned char tx0ie : 1;
	unsigned char rx1ie : 1;
	unsigned char rx0ie : 1;
} trx_int_enable_;	// CANINTE Register (Interrupt Enable)

/* Data Structure for complete Status of the MCP2515 */
typedef struct
{
	trx_int_status_	int_status;
	trx_rx_status_	rx_status;
	trx_error_		error;
	trx_config_		config;
	trx_int_flags_	int_flags;
	trx_int_enable_	int_enable;

	unsigned char conf[3];			// Configuration Registers
	unsigned char rec;				// Receive Error Counter
	unsigned char tec;				// Transmit Error Counter
} trx_status_;

#pragma pack(pop)

/* Initialisation */
int TRX_init(unsigned int baudrate);
unsigned int TRX_reset(void);
int TRX_close(void);

int TRX_status(trx_status_ *status);	// Fill the trx_status structure with all the interesting bits!

/* Basic SPI Communication Functions */
unsigned int TRX_write_register(unsigned char address, unsigned int size, unsigned char *data);
unsigned int TRX_read_register(unsigned char address, unsigned int size, unsigned char *data);
unsigned int TRX_bit_modify(unsigned char address, unsigned char mask, unsigned char data);

/* Command Functions */
unsigned int TRX_read_rx_buffer(unsigned char buffer_num, unsigned int *id, unsigned char *ext, unsigned char *rtr, unsigned char *size, unsigned char *data);
unsigned int TRX_load_tx_buffer(unsigned char buffer_num, unsigned int id, unsigned char ext, unsigned char rtr, unsigned char size, unsigned char * data);
unsigned int TRX_RTS(unsigned char buffer_num);
unsigned int TRX_read_rx_status(unsigned char *rx_status);
unsigned int TRX_read_status(unsigned char *status);

/* Configuration Functions */
int TRX_set_baudrate(unsigned int baudrate);

unsigned int TRX_set_filter_mode(unsigned char buffer_num, unsigned char filter_mode);
unsigned int TRX_select_filter(unsigned char buffer_num, unsigned char filter_num, unsigned char extended);
unsigned int TRX_set_filter_id(unsigned char filter_num, unsigned int id, unsigned char extended);
unsigned int TRX_set_mask_id(unsigned char mask_num, unsigned int id, unsigned char extended);

/* Wrapper Functions for Interrupt Handling */
inline unsigned int TRX_enable_interrupts(unsigned char interrupts);
inline unsigned int TRX_disable_interrupts(unsigned char interrupts);
inline unsigned int TRX_clear_interrupts(unsigned char interrupts);
inline unsigned int TRX_read_interrupt(unsigned char * interrupt);

/* Operation Mode Functions */
inline unsigned int TRX_read_mode(unsigned char * mode);
unsigned int TRX_switch_mode(unsigned char mode);

/* Communication Behaviour Settings */
inline unsigned int TRX_abort_transmission(void);
inline unsigned int TRX_retry_transmission(bool enable);

#endif